from typing import Dict, Any, Optional, List
from dataclasses import dataclass
from urllib.parse import urljoin
import aiohttp, asyncio
import sys, os
import shutil

if sys.platform.startswith("win"):
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

@dataclass
class ThumbnailInfo:
    url: str
    width: int
    height: int

@dataclass
class Format:
    format_id: str
    url: str
    restream: str
    width: int
    height: int
    ext: str
    codec: str
    mime_type: str
    format_note: str
    filesize: str
    bitrate: int

@dataclass
class VideoInfo:
    video_id: str
    title: str
    description: str
    view_count: str
    author: str
    duration: int
    thumbnails: List[ThumbnailInfo]
    formats: List[Format]

class ProgressBar:
    def __init__(self, total_size: int, prefix: str = 'Progress:', length: int = 50):
        self.total_size = total_size
        self.prefix = prefix
        self.length = length
        self.downloaded = 0
        self.terminal_width = shutil.get_terminal_size().columns

    def update(self, chunk_size: int):
        self.downloaded += chunk_size
        self._draw()

    def _draw(self):
        progress = self.downloaded / self.total_size
        filled_length = int(self.length * progress)
        bar = '█' * filled_length + '-' * (self.length - filled_length)
        percentage = progress * 100
        downloaded_mb = self.downloaded / 1024 / 1024
        total_mb = self.total_size / 1024 / 1024

        progress_line = f'\r{self.prefix} |{bar}| {percentage:.1f}% ({downloaded_mb:.1f}MB/{total_mb:.1f}MB)'
        
        if len(progress_line) > self.terminal_width:
            progress_line = progress_line[:self.terminal_width]

        print(progress_line, end='', flush=True)
        
        if self.downloaded >= self.total_size:
            print()

class ytdlx:
    def __init__(self, api_key: str, options: Dict[str, Any] = None, base_url: str = "https://youtube.virs.tech"):
        if not api_key:
            raise ValueError("API key is required")
        self.base_url = base_url.rstrip('/')
        self.api_key = api_key
        self.options = options
        self._session: Optional[aiohttp.ClientSession] = None

    async def __aenter__(self):
        self._session = aiohttp.ClientSession()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if self._session:
            await self._session.close()
            self._session = None

    async def _ensure_session(self):
        """Ensure we have an active session."""
        if self._session is None or self._session.closed:
            self._session = aiohttp.ClientSession()

    def _select_format(self, formats: List[Format]) -> Format:
        """Select the appropriate format based on options."""
        format_option = self.options.get("format", "bestvideo+bestaudio")
        
        def parse_format_specs(format_str: str) -> dict:
            """Parse format string to extract specifications."""
            specs = {}
            if "[" in format_str:
                specs_str = format_str[format_str.find("[")+1:format_str.find("]")]
                for spec in specs_str.split(","):
                    if "=" in spec:
                        key, value = spec.split("=")
                        specs[key] = int(value) if value.isdigit() else value
            return specs

        specs = parse_format_specs(format_option)
        base_format = format_option.split("[")[0] if "[" in format_option else format_option

        valid_formats = formats
        
        if base_format.startswith("best") or base_format.startswith("worst"):
            if "audio" in base_format:
                valid_formats = [f for f in formats if f.mime_type.startswith("audio") or f.format_note == "360p"]
            else:
                valid_formats = [f for f in formats if f.mime_type.startswith("video")]
        
        for key, value in specs.items():
            if key == "ext":
                valid_formats = [f for f in valid_formats if f.ext.lstrip('.') == value]
            elif key == "width":
                valid_formats = [f for f in valid_formats if f.width == value]
            elif key == "height":
                valid_formats = [f for f in valid_formats if f.height == value]
        
        if not valid_formats:
            raise Exception(f"No valid formats found for {format_option}")
        
        if base_format.startswith("best"):
            return max(valid_formats, key=lambda x: x.bitrate)
        elif base_format.startswith("worst"):
            return min(valid_formats, key=lambda x: x.bitrate)
        else:
            for fmt in valid_formats:
                if fmt.format_id == base_format:
                    return fmt
            raise Exception(f"Format {format_option} not found")

    async def download(self, video_id: str) -> str:
        """Download video using specified options."""
        info = await self.extract_info(video_id)
        selected_format = self._select_format(info.formats)
        print(f"Downloading: {info.title}")
        print(f"Format: {selected_format.format_note} ({selected_format.ext.lstrip('.')})")
        
        output_path = os.path.abspath(
            self.options["outtmpl"]
            .replace('%(id)s', video_id)
            .replace('%(ext)s', selected_format.ext.lstrip('.'))
        )
        
        os.makedirs(os.path.dirname(os.path.abspath(output_path)), exist_ok=True)
        
        try:
            async with self._session.get(selected_format.restream) as response:
                response.raise_for_status()
                total_size = int(response.headers.get('content-length', 0))
                progress_bar = ProgressBar(total_size, prefix='Downloading:')
                with open(output_path, 'wb') as f:
                    async for chunk in response.content.iter_chunked(8192):
                        f.write(chunk)
                        progress_bar.update(len(chunk))
            return output_path
            
        except Exception as e:
            if os.path.exists(output_path):
                os.remove(output_path)
            raise Exception(f"Download failed: {str(e)}")

    async def extract_info(self, video_id: str) -> VideoInfo:
        await self._ensure_session()
        url = f"{self.base_url}/vi/{video_id}"
        params = {'key': self.api_key}
        
        try:
            async with self._session.get(url, params=params) as response:
                if response.status != 200:
                    error_text = await response.text()
                    raise Exception(f"API request failed with status {response.status}: {error_text}")
                
                data = await response.json()
                if not data.get('success'):
                    raise Exception("API request was not successful")
                
                video_details = data['video_details']
                thumbnails = [
                    ThumbnailInfo(
                        url=thumb['url'],
                        width=thumb['width'],
                        height=thumb['height']
                    )
                    for thumb in video_details['thumbnail']['thumbnails']
                ]
                
                formats = [
                    Format(
                        format_id=str(fmt['format_id']),
                        url=fmt['url'],
                        restream=fmt['restream'],
                        width=fmt['width'],
                        height=fmt['height'],
                        ext=fmt['ext'],
                        codec=fmt['codec'],
                        mime_type=fmt['mime_type'],
                        format_note=fmt['format_note'],
                        filesize=fmt['filesize'],
                        bitrate=fmt['bitrate']
                    )
                    for fmt in data['streaming_data']
                ]
                
                return VideoInfo(
                    video_id=video_details['videoId'],
                    title=video_details['title'],
                    description=video_details['shortDescription'],
                    view_count=video_details['viewCount'],
                    author=video_details['author'],
                    duration=int(video_details['lengthSeconds']),
                    thumbnails=thumbnails,
                    formats=formats
                )
        except aiohttp.ClientError as e:
            raise Exception(f"Network error: {str(e)}")
        except KeyError as e:
            raise Exception(f"Unexpected API response format: missing key {str(e)}")
        except Exception as e:
            raise Exception(f"Error fetching video info: {str(e)}")